@@ -23,17 +23,19 @@
 #include <stdlib.h>
 
 #include "girmodule.h"
+#include "gitypelib-internal.h"
 #include "girnode.h"
 
 #define ALIGN_VALUE(this, boundary) \
   (( ((unsigned long)(this)) + (((unsigned long)(boundary)) -1)) & (~(((unsigned long)(boundary))-1)))
 
+#define NUM_SECTIONS 2
 
 GIrModule *
-g_ir_module_new (const gchar *name,
-		 const gchar *version,
-		 const gchar *shared_library,
-		 const gchar *c_prefix)
+_g_ir_module_new (const gchar *name,
+		  const gchar *version,
+		  const gchar *shared_library,
+		  const gchar *c_prefix)
 {
   GIrModule *module;
 
@@ -56,14 +58,14 @@ g_ir_module_new (const gchar *name,
 }
 
 void
-g_ir_module_free (GIrModule *module)
+_g_ir_module_free (GIrModule *module)
 {
   GList *e;
 
   g_free (module->name);
 
   for (e = module->entries; e; e = e->next)
-    g_ir_node_free ((GIrNode *)e->data);
+    _g_ir_node_free ((GIrNode *)e->data);
 
   g_list_free (module->entries);
   /* Don't free dependencies, we inherit that from the parser */
@@ -77,8 +79,8 @@ g_ir_module_free (GIrModule *module)
 }
 
 /**
- * g_ir_module_fatal:
- * @module: Current module
+ * _g_ir_module_fatal:
+ * @build: Current build
  * @line: Origin line number, or 0 if unknown
  * @msg: printf-format string
  * @args: Remaining arguments
@@ -86,12 +88,14 @@ g_ir_module_free (GIrModule *module)
  * Report a fatal error, then exit.
  */
 void
-g_ir_module_fatal (GIrModule  *module,
-                   guint       line,
-                   const char *msg,
-                   ...)
+_g_ir_module_fatal (GIrTypelibBuild  *build,
+		    guint       line,
+		    const char *msg,
+		    ...)
 {
+  GString *context;
   char *formatted;
+  GList *link;
 
   va_list args;
 
@@ -99,10 +103,27 @@ g_ir_module_fatal (GIrModule  *module,
 
   formatted = g_strdup_vprintf (msg, args);
 
-  if (line)
-    g_printerr ("%s-%s.gir:%d: error: %s\n", module->name, module->version, line, formatted);
-  else
-    g_printerr ("%s-%s.gir: error: %s\n", module->name, module->version, formatted);
+  context = g_string_new ("");
+  if (line > 0)
+    g_string_append_printf (context, "%d: ", line);
+  if (build->stack)
+    g_string_append (context, "In ");
+  for (link = g_list_last (build->stack); link; link = link->prev)
+    {
+      GIrNode *node = link->data;
+      const char *name = node->name;
+      if (name)
+	g_string_append (context, name);
+      if (link->prev)
+	g_string_append (context, ".");
+    }
+  if (build->stack)
+    g_string_append (context, ": ");
+
+  g_printerr ("%s-%s.gir:%serror: %s\n", build->module->name, 
+	      build->module->version,
+	      context->str, formatted);
+  g_string_free (context, TRUE);
 
   exit (1);
 
@@ -130,8 +151,8 @@ add_disguised_structure_foreach (gpointer key,
 }
 
 void
-g_ir_module_add_include_module (GIrModule  *module,
-				GIrModule  *include_module)
+_g_ir_module_add_include_module (GIrModule  *module,
+				 GIrModule  *include_module)
 {
   module->include_modules = g_list_prepend (module->include_modules,
 					    include_module);
@@ -165,8 +186,8 @@ write_attribute (gpointer key, gpointer value, gpointer datap)
   *(data->offset) += sizeof (AttributeBlob);
 
   blob->offset = data->node->offset;
-  blob->name = write_string ((const char*) key, data->strings, data->databuf, data->offset2);
-  blob->value = write_string ((const char*) value, data->strings, data->databuf, data->offset2);
+  blob->name = _g_ir_write_string ((const char*) key, data->strings, data->databuf, data->offset2);
+  blob->value = _g_ir_write_string ((const char*) value, data->strings, data->databuf, data->offset2);
 
   data->count++;
 }
@@ -201,13 +222,79 @@ node_cmp_offset_func (gconstpointer a,
   return na->offset - nb->offset;
 }
 
+static void
+alloc_section (guint8 *data, SectionType section_id, guint32 offset)
+{
+  int i;
+  Header *header = (Header*)data;
+  Section *section_data = (Section*)&data[header->sections];
 
-GTypelib *
-g_ir_module_build_typelib (GIrModule  *module,
-			     GList       *modules)
+  g_assert (section_id != GI_SECTION_END);
+
+  for (i = 0; i < NUM_SECTIONS; i++)
+    {
+      if (section_data->id == GI_SECTION_END)
+	{
+	  section_data->id = section_id;
+	  section_data->offset = offset;
+	  return;
+	}
+      section_data++;
+    }
+  g_assert_not_reached ();
+}
+
+static guint8*
+add_directory_index_section (guint8 *data, GIrModule *module, guint32 *offset2)
+{
+  DirEntry *entry;
+  Header *header = (Header*)data;
+  GITypelibHashBuilder *dirindex_builder;
+  guint i, n_interfaces;
+  guint16 required_size;
+  guint32 new_offset;
+
+  dirindex_builder = _gi_typelib_hash_builder_new ();
+
+  n_interfaces = ((Header *)data)->n_local_entries;
+
+  for (i = 0; i < n_interfaces; i++)
+    {
+      entry = (DirEntry *)&data[header->directory + (i * header->entry_blob_size)];
+      const char *str = (const char *) (&data[entry->name]);
+      _gi_typelib_hash_builder_add_string (dirindex_builder, str, i);
+    }
+
+  if (!_gi_typelib_hash_builder_prepare (dirindex_builder))
+    {
+      /* This happens if CMPH couldn't create a perfect hash.  So
+       * we just punt and leave no directory index section.
+       */
+      _gi_typelib_hash_builder_destroy (dirindex_builder);
+      return data;
+    }
+
+  alloc_section (data, GI_SECTION_DIRECTORY_INDEX, *offset2);
+
+  required_size = _gi_typelib_hash_builder_get_buffer_size (dirindex_builder);
+
+  new_offset = *offset2 + ALIGN_VALUE (required_size, 4);
+
+  data = g_realloc (data, new_offset);
+
+  _gi_typelib_hash_builder_pack (dirindex_builder, ((guint8*)data) + *offset2, required_size);
+
+  *offset2 = new_offset;
+
+  _gi_typelib_hash_builder_destroy (dirindex_builder);
+  return data;
+}
+
+GITypelib *
+_g_ir_module_build_typelib (GIrModule  *module)
 {
   GError *error = NULL;
-  GTypelib *typelib;
+  GITypelib *typelib;
   gsize length;
   guint i;
   GList *e;
@@ -223,6 +310,7 @@ g_ir_module_build_typelib (GIrModule  *module,
   GList *nodes_with_attributes;
   char *dependencies;
   guchar *data;
+  Section *section;
 
   header_size = ALIGN_VALUE (sizeof (Header), 4);
   n_local_entries = g_list_length (module->entries);
@@ -268,7 +356,7 @@ g_ir_module_build_typelib (GIrModule  *module,
     {
       GIrNode *node = e->data;
 
-      size += g_ir_node_get_full_size (node);
+      size += _g_ir_node_get_full_size (node);
 
       /* Also reset the offset here */
       node->offset = 0;
@@ -283,6 +371,8 @@ g_ir_module_build_typelib (GIrModule  *module,
   if (module->c_prefix != NULL)
     size += ALIGN_VALUE (strlen (module->c_prefix) + 1, 4);
 
+  size += sizeof (Section) * NUM_SECTIONS;
+
   g_message ("allocating %d bytes (%d header, %d directory, %d entries)\n",
 	  size, header_size, dir_size, size - header_size - dir_size);
 
@@ -291,7 +381,7 @@ g_ir_module_build_typelib (GIrModule  *module,
   /* fill in header */
   header = (Header *)data;
   memcpy (header, G_IR_MAGIC, 16);
-  header->major_version = 3;
+  header->major_version = 4;
   header->minor_version = 0;
   header->reserved = 0;
   header->n_entries = n_entries;
@@ -302,20 +392,19 @@ g_ir_module_build_typelib (GIrModule  *module,
    * the size calculations above.
    */
   if (dependencies != NULL)
-    header->dependencies = write_string (dependencies, strings, data, &header_size);
+    header->dependencies = _g_ir_write_string (dependencies, strings, data, &header_size);
   else
     header->dependencies = 0;
   header->size = 0; /* filled in later */
-  header->namespace = write_string (module->name, strings, data, &header_size);
-  header->nsversion = write_string (module->version, strings, data, &header_size);
+  header->namespace = _g_ir_write_string (module->name, strings, data, &header_size);
+  header->nsversion = _g_ir_write_string (module->version, strings, data, &header_size);
   header->shared_library = (module->shared_library?
-                             write_string (module->shared_library, strings, data, &header_size)
+                             _g_ir_write_string (module->shared_library, strings, data, &header_size)
                              : 0);
   if (module->c_prefix != NULL)
-    header->c_prefix = write_string (module->c_prefix, strings, data, &header_size);
+    header->c_prefix = _g_ir_write_string (module->c_prefix, strings, data, &header_size);
   else
     header->c_prefix = 0;
-  header->directory = ALIGN_VALUE (header_size, 4);
   header->entry_blob_size = sizeof (DirEntry);
   header->function_blob_size = sizeof (FunctionBlob);
   header->callback_blob_size = sizeof (CallbackBlob);
@@ -335,10 +424,26 @@ g_ir_module_build_typelib (GIrModule  *module,
   header->interface_blob_size = sizeof (InterfaceBlob);
   header->union_blob_size = sizeof (UnionBlob);
 
+  offset2 = ALIGN_VALUE (header_size, 4);
+  header->sections = offset2;
+
+  /* Initialize all the sections to _END/0; we fill them in later using
+   * alloc_section().  (Right now there's just the directory index
+   * though, note)
+   */
+  for (i = 0; i < NUM_SECTIONS; i++)
+    {
+      section = (Section*) &data[offset2];
+      section->id = GI_SECTION_END;
+      section->offset = 0;
+      offset2 += sizeof(Section);
+    }
+  header->directory = offset2;
+
   /* fill in directory and content */
   entry = (DirEntry *)&data[header->directory];
 
-  offset2 = header->directory + dir_size;
+  offset2 += dir_size;
 
   for (e = module->entries, i = 0; e; e = e->next, i++)
     {
@@ -380,33 +485,33 @@ g_ir_module_build_typelib (GIrModule  *module,
 
 	  entry->blob_type = 0;
 	  entry->local = FALSE;
-	  entry->offset = write_string (namespace, strings, data, &offset2);
-	  entry->name = write_string (node->name, strings, data, &offset2);
+	  entry->offset = _g_ir_write_string (namespace, strings, data, &offset2);
+	  entry->name = _g_ir_write_string (node->name, strings, data, &offset2);
 	}
       else
 	{
 	  old_offset = offset;
-	  offset2 = offset + g_ir_node_get_size (node);
+	  offset2 = offset + _g_ir_node_get_size (node);
 
 	  entry->blob_type = node->type;
 	  entry->local = TRUE;
 	  entry->offset = offset;
-	  entry->name = write_string (node->name, strings, data, &offset2);
+	  entry->name = _g_ir_write_string (node->name, strings, data, &offset2);
 
+	  memset (&build, 0, sizeof (build));
 	  build.module = module;
-	  build.modules = modules;
 	  build.strings = strings;
 	  build.types = types;
 	  build.nodes_with_attributes = nodes_with_attributes;
 	  build.n_attributes = header->n_attributes;
 	  build.data = data;
-	  g_ir_node_build_typelib (node, NULL, &build, &offset, &offset2);
+	  _g_ir_node_build_typelib (node, NULL, &build, &offset, &offset2);
 
 	  nodes_with_attributes = build.nodes_with_attributes;
 	  header->n_attributes = build.n_attributes;
 
-	  if (offset2 > old_offset + g_ir_node_get_full_size (node))
-	    g_error ("left a hole of %d bytes\n", offset2 - old_offset - g_ir_node_get_full_size (node));
+	  if (offset2 > old_offset + _g_ir_node_get_full_size (node))
+	    g_error ("left a hole of %d bytes\n", offset2 - old_offset - _g_ir_node_get_full_size (node));
 	}
 
       entry++;
@@ -434,6 +539,10 @@ g_ir_module_build_typelib (GIrModule  *module,
 
   data = g_realloc (data, offset2);
   header = (Header*) data;
+
+  data = add_directory_index_section (data, module, &offset2);
+  header = (Header *)data;
+
   length = header->size = offset2;
   typelib = g_typelib_new_from_memory (data, length, &error);
   if (!typelib)
