@@ -90,6 +90,10 @@ def _get_option_parser():
     parser.add_option("", "--no-libtool",
                       action="store_true", dest="nolibtool", default=False,
                       help="do not use libtool")
+    parser.add_option("", "--external-library",
+                      action="store_true", dest="external_library", default=False,
+                      help=("""If true, the library is located on the system,""" +
+                            """not in the current directory"""))
     parser.add_option("-l", "--library",
                       action="append", dest="libraries", default=[],
                       help="libraries of this unit")
@@ -215,16 +219,6 @@ def process_packages(options, packages):
     options.cpp_defines.extend(pkg_options.cpp_defines)
     options.cpp_undefines.extend(pkg_options.cpp_undefines)
 
-    args = ['pkg-config', '--libs-only-L']
-    args.extend(packages)
-    output = subprocess.Popen(args,
-                              stdout=subprocess.PIPE).communicate()[0]
-    if output is None:
-        return 1
-    filtered_output = list(process_options(output, options_whitelist))
-    pkg_options, unused = parser.parse_args(filtered_output)
-    options.library_paths.extend(pkg_options.library_paths)
-
 def extract_filenames(args):
     filenames = []
     for arg in args:
@@ -415,9 +409,13 @@ def scanner_main(args):
     final = IntrospectablePass(transformer, blocks)
     final.validate()
 
-    if options.warn_fatal and logger.did_warn():
+    warning_count = logger.get_warning_count()
+    if options.warn_fatal and warning_count > 0:
         message.fatal("warnings configured as fatal")
         return 1
+    elif warning_count > 0 and options.warn_all is False:
+        print ("g-ir-scanner: %s: warning: %d warnings suppressed (use --warn-all to see them)"
+               % (transformer.namespace.name, warning_count, ))
 
     # Write out AST
     if options.packages_export:
